require "numru/gphys/coordmapping"
require "numru/gphys/gphys"
begin
  require "numru/ssl2"
  HAVE_NUMRU_SSL2 = true
rescue LoadError
  HAVE_NUMRU_SSL2 = false
end

=begin
=class NumRu::GPhys

Additional methods regarding coordinate transformation.

== Methods

---coordtransform( coordmapping, axes_to, *dims )

     Coordinate transformation with ((|coordmapping|)) into ((|axes_to|)).

     ARGUMENTS
     * ((|coordmapping|)) (CoordMapping) : relation between the new and
       original coordinate systems. Mapping is defined from the new one to
       the orinal one. If the rank of the mapping is smaller than 
       the rank of self, ((|dims|)) must be used to specify the
       correspondence. e.g., if the rank of the mapping is 2 and that 
       of self is 3 and the mapping is regarding the first 2 dimensions
       of the three, ((|dims|)) must be [0,1].
     * ((|axes_to|)) (Array of Axis) : grid in the new coordinate system.
       Its length must be the same as the rank of ((|coordmapping|))
     * ((|dims|)) (integers) : Specifies the dimensions to which 
       ((|coordmapping|)) is applied. Needed if 
       (({self.rank!=coordmapping.rank})) (neglected otherwise). 
       The number of integers must agree with the rank of the mapping.

=end

module NumRu
  class GPhys

    def coordtransform( coordmapping, axes_to, *dims )

      rankmp = coordmapping.rank

      #< check arguments >
      if axes_to.length != rankmp
        raise ArgumentError,
          "length of axes_to must be equal to the rank of coordmapping"
      end
      if self.rank == rankmp
        dims = (0...rankmp).collect{|i| i}
      elsif self.rank < rankmp
        raise ArgumentError,"rank of coordmapping is greater than self.rank"
      elsif dims.length != rankmp
        raise ArguemntError,
          "# of dimensions speficied is not equal to the rank of coordmapping"
      elsif dims != dims.sort
        raise ArguementErroor,"dims must be in the increasing order"
      end

      #< get grid points >
      vt = coordmapping.map_grid( *dims.collect{|d| axes_to[d].pos.val} )
      x = dims.collect{|d| self.grid.axis(d).pos.val}
      p x
      #< prepare the output object >
      axes = (0...self.rank).collect{|i| grid.axis(i)}
      dims.each_with_index{|d,j| axes[d]=axes_to[j]}
      grid_to = Grid.new( *axes )
      vnew = VArray.new( NArray.new( self.data.ntype, *grid_to.shape ),
                        self.data, self.name )

      #< do interpolation (so far only 2D is supported) >
      case dims.length
      when 2
        if !HAVE_NUMRU_SSL2
#         raise "Sorry, so far I need SSL2 (ruby-ssl2)"
          self.each_subary_at_dims_with_index( *dims ){ |fxy,idx|

            lx = vt[0].shape[0]
            x0 = NArray.int(lx)
            x1 = NArray.int(lx)
            wx = NArray.float(lx)
            for i in 0..lx-1
              index = (x[0]<=vt[0][i,0]).count_true
#              x1[i] = (x[0]<=vt[0][i,0]).count_true
#              x0[i] = x1[i] - 1
#              if x0[i]==-1
              if index==0
                x0[i] = x[0].sort_index[0]
                x1[i] = x0[i]
                wx[i] = 0.0
#              elsif x1[i]==x[0].length
              elsif index==x[0].length
                x0[i] = x[0].sort_index[-1]
                x1[i] = x0[i]
                wx[i] = 1.0
              else
                x0[i] = x[0].sort_index[index-1]
                x1[i] = x[0].sort_index[index]
                wx[i] = (vt[0][i,0]-x[0][x0[i]])/(x[0][x1[i]]-x[0][x0[i]])
              end
            end

            ly = vt[1].shape[1]
            y0 = NArray.int(ly)
            y1 = NArray.int(ly)
            wy = NArray.float(ly)
            for j in 0..ly-1
              index = (x[1]<=vt[1][0,j]).count_true
#              y1[j] = (x[1]<=vt[1][0,j]).count_true
#              y0[j] = y1[j] - 1
              if index==0
                y0[j] = x[1].sort_index[0]
                y1[j] = y0[j]
                wy[j] = 0.0
              elsif index==x[1].length
                y0[j] = x[1].sort_index[-1]
                y1[j] = y0[j]
                wy[j] = 1.0
              else
                y0[j] = x[1].sort_index[index-1]
                y1[j] = x[1].sort_index[index]
                wy[j] = (vt[1][0,j]-x[1][y0[j]])/(x[1][y1[j]]-x[1][y0[j]])
              end
            end

#            f = ( fxy[x0,y0]*(1-wx) + fxy[x1,y0]*wx )*(1-wy) +
#                ( fxy[x0,y1]*(1-wx) + fxy[x1,y1]*wx )*wy
            f = ( fxy.data.val[x0,y0]*(1-wx) + fxy.data.val[x1,y0]*wx ) * 
                 ((1-wy).newdim(0)) +
               ( fxy.data.val[x0,y1]*(1-wx) + fxy.data.val[x1,y1]*wx ) * 
                 (wy.newdim(0))

            if(idx==false)
              vnew[] = f
            else
              vnew[*idx] = f
            end

#            for j in 0..ly-1
#              for i in 0..lx-1
#                vnew[i,j] = ( fxy.data.val[x0[i],y0[j]]*(1-wx[i]) + 
#                              fxy.data.val[x1[i],y0[j]]*wx[i] )*(1-wy[j]) +
#                            ( fxy.data.val[x0[i],y1[j]]*(1-wx[i]) + 
#                              fxy.data.val[x1[i],y1[j]]*wx[i] )*wy[j]
#              end
#            end

          }
#       end
        else
          ix=iy=0
          m=3
          self.each_subary_at_dims_with_index( *dims ){ |fxy,idx|
            c,xt = SSL2.bicd3(x[0],x[1],fxy.val,m)
            begin
              ix,iy,f = SSL2.bifd3(x[0],x[1],m,c,xt,0,vt[0],ix,0,vt[1],iy)
            rescue
              $stderr.print "Interpolation into", vt[0].inspect, vt[1].inspect
              raise $!
            end
            vnew[*idx] = f
          }
        end
      else
        raise "Sorry, #{v.length}D interpolation is yet to be supported"
      end

      #< finish >
      GPhys.new( grid_to, vnew )
    end

  end
end

########################################

if __FILE__ == $0
  include NumRu
  include NMath

  #< make a GPhys >

  puts "** preparation **"

  nx=ny=10
  nz=3
  xv = VArray.new( xx=NArray.sfloat(nx).indgen!.mul!(4*PI/(nx-1)) ).rename("x")
  yv = VArray.new( yy=NArray.sfloat(ny).indgen!.mul!(4*PI/(ny-1)) ).rename("y")
  zv = VArray.new( NArray.sfloat(nz).indgen! ).rename("z")
  xax = Axis.new.set_pos(xv)
  yax = Axis.new.set_pos(yv)
  zax = Axis.new.set_pos(zv)
  grid = Grid.new(xax,yax,zax)
  fxy = sin(yy).newdim(0,1) + NArray.sfloat(nx).newdim(1,1) +
                              NArray.sfloat(nz).indgen!.newdim(0,0)
  p fxy.shape, fxy
  z = VArray.new( fxy )
  gp = GPhys.new( grid, z )


  #< make the new coordinate  >

  puts "** transformation **"

  theta = PI/6
  factor = NMatrix[ [cos(theta), -sin(theta)],
                    [sin(theta), cos(theta) ] ]
  offset = NVector[ 0.0, 0.0 ]
  coordmapping = LinearCoordMapping.new(offset, factor)

  axes_to = [ Axis.new.set_pos( xv[0..(nx/2)] + 1.5*PI ), 
              Axis.new.set_pos( yv[0..(ny/2)] ) ]

  gprot = gp.coordtransform( coordmapping, axes_to, 0, 1 )
  p gprot.val

end

